/*****************************************************************************

	[book.cpp] Book class implementation
	By Jeff Leung

*****************************************************************************/
#include "book.h"
#include "client.h"
#include "misc.h"

//_[Book::Book()]_____________________________________________________________
//
// DESCRIPTION:
//  Default constructor, sets borrower pointer to NULL.
//
// PARAMETERS:
//  None.
//
Book::Book()
{
	pBorrower = NULL;
}

//_[Book::~Book()]____________________________________________________________
//
// DESCRIPTION:
//  Destructor.
//
Book::~Book()
{
}

//_[Book::getIDProp()]________________________________________________________
//
// See Item::getIDProp()
//
string Book::getIDProp(void)
{
	string sIDProp = "ISBN";
	return sIDProp;
}

//_[Book::getName()]__________________________________________________________
//
// See Item::getName()
//
string Book::getName(unsigned int iProp)
{
	string sName;

	switch (iProp)
	{
		case 0:		sName = "ISBN";			break;
		case 1:		sName = "TITLE";		break;
		case 2:		sName = "AUTHOR";		break;
		case 3:		sName = "PUBLISHER";	break;
		case 4:		sName = "OUTDATE";		break;
		case 5:		sName = "DUEDATE";		break;
		case 6:		sName = "BORROWER";		break;
	}

	return sName;
}

//_[Book::getValue()]_________________________________________________________
//
// See Item::getValue()
//
string Book::getValue(string &sProp)
{
	if (sProp == "ISBN")
		return sISBN;
	else if (sProp == "AUTHOR")
		return sAuthor;
	else if (sProp == "TITLE")
		return sTitle;
	else if (sProp == "PUBLISHER")
		return sPublisher;
	else if (sProp == "OUTDATE")
		return sOutDate;
	else if (sProp == "DUEDATE")
		return sDueDate;
	else if (sProp == "BORROWER")
		return sBorrower;
}

//_[Book::getDesc()]__________________________________________________________
//
// See Item::getDesc()
//
string Book::getDesc(string &sProp)
{
	string sDesc;

	if (sProp == "ISBN")
		sDesc = "The unique identifying number for a book.";
	else if (sProp == "AUTHOR")
		sDesc = "The writer or main contributor for the book.";
	else if (sProp == "TITLE")
		sDesc = "The title of the book.";
	else if (sProp == "PUBLISHER")
		sDesc = "The publisher of the book.";
	else if (sProp == "OUTDATE")
		sDesc = "The date the book was borrowed (leave blank if not borrowed).";
	else if (sProp == "DUEDATE")
		sDesc = "The date the book is due to be returned (leave blank if not borrowed).";
	else if (sProp == "BORROWER")
		sDesc = "ID of client borrowing book (leave blank if not borrowed).";

	return sDesc;
}

//_[Book::setProp()]__________________________________________________________
//
// See Item::setProp()
//
Item::return_value Book::setProp(string &sProp, string &sValue)
{
	if (sProp == "ISBN")
	{
		sISBN = sValue;
	}
	else if (sProp == "AUTHOR")
	{
		sAuthor = sValue;
	}
	else if (sProp == "TITLE")
	{
		sTitle = sValue;
	}
	else if (sProp == "PUBLISHER")
	{
		sPublisher = sValue;
	}
	else if (sProp == "OUTDATE")
	{
		if (sValue == "" || isDateValid(sValue))
			sOutDate = sValue;
		else
			return Item::BadValue;
	}
	else if (sProp == "DUEDATE")
	{
		if (sValue == "" || isDateValid(sValue))
			sDueDate = sValue;
		else
			return Item::BadValue;
	}
	else if (sProp == "BORROWER")
	{
		sBorrower = sValue;
	}

	return Item::Okay;
}

//_[Book::numProp()]__________________________________________________________
//
// See Item::getName()
//
unsigned int Book::numProps(void)
{
	return 7;
}

//_[Book::isBorrowed()]_______________________________________________________
//
// DESCRIPTION:
//  Determines if the book is currently being borrowed.
//
// PARAMETERS:
//  None.
//
// RETURN VALUE (bool):
//  Returns true if the book is being borrowed, false otherwise.
//
bool Book::isBorrowed(void)
{
	return (pBorrower==NULL)?false:true;
}

//_[Book::isDue()]____________________________________________________________
//
// DESCRIPTION:
//  Determines if the book is due (the due date is past).
//
// PARAMETERS:
//  None.
//
// RETURN VALUE (bool):
//  Returns true if the book is due, false otherwise.
//
bool Book::isDue(void)
{
	if (!isBorrowed())
		return false;

	return hasDatePast(sDueDate);
}

//_[Book::SetBorrower()]______________________________________________________
//
// DESCRIPTION:
//  Sets the borrower of this book to a specified client.  This should be
//  called BY the client borrowing the book.  The book should never call this
//  directly because the client will not be notified.
//
// PARAMETERS:
//  pClient ............... client who will be borrowing book
//  sDate ................. date the book is due to be returned
//
// RETURN VALUE (bool):
//  Returns true on success, false on failure.
//
bool Book::SetBorrower(Item *pClient, string &sDate)
{
	if (isBorrowed())
		return false;

	string sClientID = pClient->getIDProp();

	sOutDate	= getTodaysDate();
	sDueDate	= sDate;
	sBorrower	= pClient->getValue(sClientID);
	pBorrower	= pClient;

	return true;
}

//_[Book::ClearBorrower()]____________________________________________________
//
// DESCRIPTION:
//  Clears the borrower information stored.  This should be called by the
//  client that was borrowing the book.  That way the link is clearly
//  destroyed between the two.  The book should never call this directly
//  because the client will not be notified.
//
// PARAMETERS:
//  None.
//
// RETURN VALUE (void):
//  None.
//
void Book::ClearBorrower(void)
{
	sOutDate	= "";
	sDueDate	= "";
	sBorrower	= "";
	pBorrower	= NULL;
}

//_[Book::InitBorrower()]_____________________________________________________
//
// DESCRIPTION:
//  Restores the previous link between the borrowing client and this book, as
//  when data is loaded from a file.  This restoration is initiated by the 
//  book and the client is called to restore the link on it's end as well.
//
// PARAMETERS:
//  pClient ............... borrowing client
//
// RETURN VALUE (bool):
//  Returns true on success, false on failure.
//
bool Book::InitBorrower(Item *pClient)
{
	string sClientID = pClient->getIDProp();

	if (pClient->getValue(sClientID) != sBorrower)
		return false;

	pBorrower = pClient;

	reinterpret_cast<Client*>(pClient)->InitBorrowed(reinterpret_cast<Book*>(this));

	return true;
}