// =============================================================================
// Simple Dictionary
// By <name>
//
// Reads in a data file from the name specified and performs basic dictionary
// operations based on the instructions there: add, look-up, dump, and exit.
// =============================================================================

//
// include header files
//
#include <iostream>	// for console i/o (cin/cout)
#include <fstream>  // for loading data from file
#include <iomanip>	// i/o manipulation (setw)
#include <string>	// for all string variables (string)

// use standard namespace
using namespace std;

//
// Single entry in dictionary.
//
struct Entry
{
	string Word;	// entry word
	string Def;		// definition of word
};

//
// Function Prototypes
//
bool ParseLine(ifstream &InFile, string &Cmd, string &Word, string &Def);
bool AddWord(string &Word, string &Def, Entry Dictionary[], int &Count,
				int MaxLen);
bool LookUpWord(string &Word, Entry Dictionary[], int Count);
bool DumpWords(Entry Dictionary[], int Count);

// =============================================================================
// Program entry point.
// =============================================================================
int main()
{
	//
	// program variables
	//
	const int MaxEntries = 25;		// maximum number of entries in dictionary
	Entry Dictionary[MaxEntries];	// array of all dictionary entries
	int Count = 0;					// count of entries in dictionary
	string DataFileName;			// name of data file
	ifstream DataFile;				// stream for input data file
	string Command;					// single command name (e.g. "ADD", etc.)
	string Word;					// word parameter for command
	string Def;						// definition parameter for command

	//
	// let user specify data file to use
	//
	cout << "Please enter data file name: ";
	cin >> DataFileName;

	//
	// open the data file
	//
	DataFile.open(DataFileName.c_str());

	if (DataFile.fail())
	{
		cout << "Failed to open \"" << DataFileName << "\"!" << endl;
		exit(1);
	}

	//
	// read commands from data file until eof is reached
	//
	while (ParseLine(DataFile, Command, Word, Def))
	{
		// output a blank line before command
		cout << endl << "Read: " << Command;

		//- [ADD command] --------------------------------------------------
		if (Command == "ADD")
		{
			// output word and definition)
			cout << " (" << Word << "=\"" << Def << "\")" << endl;

			// attempt to add word
			AddWord(Word, Def, Dictionary, Count, MaxEntries);
		}
		//- [LOOKUP command] -----------------------------------------------
		else if (Command == "LOOKUP")
		{
			// read in word to look-up
			cout << " (" << Word << ")" << endl;

			// attempt to look-up word
			LookUpWord(Word, Dictionary, Count);
		}
		//- [DUMP command] -------------------------------------------------
		else if (Command == "DUMP")
		{
			// end line (nothing else to show for this command)
			cout << endl;

			// attempt to dump all dictionary entries
			DumpWords(Dictionary, Count);
		}
		//- [EXIT command] -------------------------------------------------
		else if (Command == "EXIT")
		{
			// output info to user
			cout << endl << "Exiting program.  Good bye" << endl;

			// break from data file loop ... no need to skip remainder of
			// line because we're exiting and the stream will be closing
			break;
		}
		//- [unknown] ------------------------------------------------------
		else
		{
			// output error to user
			cout << " (Invalid or unknown command!)" << endl;
		}

	} // while (!eof)

	//
	// close data file
	//
	DataFile.close();

	return 0;
}

// =============================================================================
// Reads a line from an input file and parses it into three pieces: a command,
// word parameter (optional), and definition parameter (optional).  If any of
// the latter two parameters do not exist, they are set to blank.  If a line
// could not be read from the input file OR if the parsing fails, the function
// returns false.
// =============================================================================
bool ParseLine(ifstream &InFile, string &Cmd, string &Word, string &Def)
{
	bool fReturn = true;	// return value of function, true until error
	string Line;			// contains an entire line from the file

	// If the file pointer for the input file is already at the end or if the
	// stream is errored, set return value to false and exit.
	if (InFile.eof() || InFile.fail())
	{
		fReturn = false;
	}
	// Otherwise if reading a line from the file into our buffer 'Line' is
	// successful ...
	else if (getline(InFile, Line))
	{
		// Position variable used to find location of spaces for the purpose
		// of trimming.
		string::size_type pos = 0;

		// reset word and definition to blank in case they aren't parsed
		Word = Def = "";

		// parse trimmed command
		pos = Line.find(" ", 0);
		Cmd = Line.substr(0, pos);

		// parse word parameter (if any)
		if (Line.length() >= 7)
		{
			pos = Line.find(" ", 7);
			Word = Line.substr(7, pos - 7);
		}

		// parse remainder of line (if any) into definition parameter; this
		// doesn't need trimming because it extends to the end of the line
		if (Line.length() >= 23)
		{
			Def = Line.substr(23, Line.length() - 23);
		}
	}
	// Else if reading a line from the file failed, set the return value to
	// false and exit.
	else
	{
		fReturn = false;
	}

	return fReturn;
}

// =============================================================================
// Adds a word to the dictionary.  Words are added in ascending alphabetical
// order to the current array of entries.
// =============================================================================
bool AddWord(string &Word, string &Def, Entry Dictionary[], int &Count,
				int MaxLen)
{
	bool fReturn = true; // true until error

	// Make sure the maximum size (MaxLen) and count of entries (Count)
	// parameters are correct:
	//
	// The count of entries mustn't be negative and mustn't be greater than the
	// maximum size (otherwise the maximum size would have already been
	// exceeded).
	//
	if (MaxLen < 1 || Count < 0 || Count > MaxLen)
	{
		cout << "Count of entries or maximum dictionary length is invalid!" << endl;
		fReturn = false;
	}
	// Else if the count is equivalent to the maximum size then there is no more
	// room for another word
	else if (Count == MaxLen)
	{
		cout << "No more room to add new words!" << endl;
		fReturn = false;
	}
	// Else if this is the first word in the dictionary, simply set the first
	// subscript as such.
	else if (Count == 0)
	{
		Dictionary[Count].Word = Word;
		Dictionary[Count].Def = Def;
	}
	// Otherwise we need to find an appropriate place for this word
	else
	{
		//
		// Loop through all subscripts until end is reached or until word is
		// found to be "lesser" (alphabetically) than the current entry word.
		// This is done so that the word will be inserted at a point that
		// preserves the ascending alphabetical order of the dictionary.
		//
		int i = 0;
		while (Word > Dictionary[i].Word && i < Count)
		{
			i++;
		}

		// If end was reached, append word to the end of the dictionary.
		if (i >= Count)
		{
			Dictionary[Count].Word = Word;
			Dictionary[Count].Def = Def;
		}
		// Otherwise check if word entry is the same (already in dictionary)
		else if (Dictionary[i].Word == Word)
		{
			cout << "Cannot add word \"" << Word
				<< "\"; it already exists in dictionary!" << endl;
			fReturn = false;
		}
		// Else the word is not in the dictionary and should be inserted at the
		// index stored in 'i'.
		else
		{
			// copy insertion position to new variable so 'i' can be used again
			// as an index in the loop below.
			int InsertPos = i;

			// shift all word entries down
			i = Count;
			while (i > InsertPos)
			{
				Dictionary[i].Word = Dictionary[i-1].Word;
				Dictionary[i].Def = Dictionary[i-1].Def;
				i--;
			}

			// add word at InsertPos
			Dictionary[InsertPos].Word = Word;
			Dictionary[InsertPos].Def = Def;
		}

	} // else (find place to insert word)

	//
	// Output success information ...
	//
	if (fReturn)
	{
		cout << "Added \"" << Word << "\" as \"" << Def << "\"."
			<< endl;

		Count ++; // increment number of entries in dictionary
	}

	return fReturn;
}

// =============================================================================
// Looks up a word in the dictionary.  Returns false if the word could not be
// found.
// =============================================================================
bool LookUpWord(string &Word, Entry Dictionary[], int Count)
{
	bool fReturn = false; // failure until word is found

	// If the dictionary has no words ...
	if (Count <= 0)
	{
		cout << "Cannot look up word; dictionary contains no entries!"
			<< endl;
	}
	// Otherwise search words of dictionary ...
	else
	{
		// Loop through all words in dictionary
		for (int i = 0; i < Count; i++)
		{
			// Found word, show definition
			if (Dictionary[i].Word == Word)
			{
				// output findings
				cout << "Found \"" << Word << "\" as \"" << Dictionary[i].Def
					<< "\"" << endl;

				// found word
				fReturn = true;
			}
		}

		// show failure
		if (fReturn == false)
		{
			cout << "Could not find \"" << Word << "\" in dictionary!" << endl;
		}

	} // else (dictionary is okay to be searched)

	return fReturn;
}

// =============================================================================
// Dumps all the entries of the dictionary to the output.
// =============================================================================
bool DumpWords(Entry Dictionary[], int Count)
{
	bool fReturn = true;

	// If the dictionary has no words ...
	if (Count <= 0)
	{
		cout << "Cannot dump entries list; dictionary contains no entries!"
			<< endl;

		fReturn = false;
	}
	// Otherwise dump words of dictionary ...
	else
	{
		//
		// output dump header
		//
		cout	<< "Words          Definitions" << endl
				<< "-------------- "
				<< "---------------------------------------------------------"
				<< endl;

		//
		// output each word and definition
		//
		for (int i = 0; i < Count; i++)
		{
			//
			// Output each word in a 15 character block that is left-aligned.
			// The definition follows without an absolute size.
			//
			cout << setiosflags(ios_base::left) << setw(15)
				<< Dictionary[i].Word << Dictionary[i].Def << endl;

		} // for (i = 0 to Count of entries)

	} // else (dictionary is okay to be "dumped")

	return fReturn;
}
