//------------------------------------------------------------------------------
// File: scores.cpp
// Desc: Source for "scores" program which loads an array of up to 50 scores
//		from a file and then presents a menu to allow the user to print the
//		scores in ascending order, calculate and print the average, calculate
//		and print the median, and calculate and print the standard deviation.
//------------------------------------------------------------------------------

//
// include headers
//
#include <iostream.h>		// i/o streaming (cout)
#include <fstream.h>		// file streaming (ifstream)
#include <math.h>			// math functions (sqrt)

//
// function prototypes
//
void SortAscending(float Source[], float Destination[], int Count);
void PrintScores(float Scores[], int ScoreCount);
float GetMean(float Scores[], int ScoreCount);
void PrintAverage(float Scores[], int ScoreCount);
void PrintMedian(float Scores[], int ScoreCount);
void PrintStdDeviation(float Scores[], int ScoreCount);

//
// globals
//
const int MaxScoreCount = 50;

//------------------------------------------------------------------------------
// Name: main()
// Desc: Program entry point.
//------------------------------------------------------------------------------
int main()
{
	float Scores[MaxScoreCount];
	int ScoreCount = 0;
	ifstream InDataFile;
	char cOption;

	//
	// load scores data file
	//
	InDataFile.open("score.dat");

	// loop and read scores until EOF is reached
	//
	// !NOTE! There cannot be any white space (spaces, tabs, new-lines, etc.)
	// following the last entry IF that last entry is the maximum.  Otherwise,
	// this logic will assume that there is another entry following that one
	// and decide that there are too many scores to store.
	//
	while (!InDataFile.eof())
	{
		// trying to read more scores than there is available space
		if (ScoreCount >= MaxScoreCount)
		{
			cout << "Too many scores in input data file!" << endl;

			// return "error" (non-zero) code (this exits the program)
			return 1;
		}

		// read next score from data file
		InDataFile >> Scores[ScoreCount];

		// increment number of scores in array
		ScoreCount++;

	} // while (!InDataFile.eof())

	// close data file
	InDataFile.close();

	//
	// loop
	//
	do
	{
		// show menu
		cout << endl
			<< "Menu:" << endl
			<< "[p] print scores in ascending order" << endl
			<< "[a] calculate and print the average value of the scores" << endl
			<< "[m] calculate and print the median value for the scores" << endl
			<< "[s] calcualte and print the standard deviation of the scores" << endl
			<< "[q] quit" << endl
			<< endl
			<< "Enter choice and press enter: ";

		// have user enter desired selection
		cin >> cOption;

		//
		// do something based on user entered selection
		//
		switch (cOption)
		{
			// Print scores in ascending order
			case 'p':
				PrintScores(Scores, ScoreCount);
				break;

			// Calculate and print the average value of the scores
			case 'a':
				PrintAverage(Scores, ScoreCount);
				break;

			// Calculate and print the median value for the scores
			case 'm':
				PrintMedian(Scores, ScoreCount);
				break;

			// Calculate and print the standard deviation of the scores
			case 's':
				PrintStdDeviation(Scores, ScoreCount);
				break;

			// Quit the program
			case 'q':
				// do nothing ... this will end the loop
				break;

			default:
				cout << "Unknown command '" << cOption << "'!" << endl;
				break;
		}

	} while (cOption != 'q');

	// exit
	return 0;
}

//------------------------------------------------------------------------------
// Name: SortAscending()
// Desc: Sorts the scores from the source array (Source[]) into ascending order
//		into the destination array (Destination[]).  The number of scores in the
//		array is specified by 'Count'.
//
//		Currently uses a simple bubble-sort whereby the source array is
//		traversed completely a number of times as is equivalent to the count.
//		On each traversal, two scores are compared and if the latter is found
//		to be "less" (should be earlier in array), the two scores are swapped.
//------------------------------------------------------------------------------
void SortAscending(float Source[], float Destination[], int Count)
{
	int Traverse, i;
	float Temp;

	// first copy the source array to the destination
	for (i = 0; i < Count; i++)
	{
		Destination[i] = Source[i];
	}

	// traverse the destination array multiple times to sort it
	for (Traverse = 0; Traverse < Count; Traverse++)
	{
		// one traversal through the array swaps each value zero or more times
		for (i = 1; i < Count; i++)
		{
			// this entry is smaller than the last, swap this subscript with
			// the previous
			if (Destination[i] < Destination[i-1])
			{
				Temp = Destination[i];
				Destination[i] = Destination[i-1];
				Destination[i-1] = Temp;
			}
		}
	}
}

//------------------------------------------------------------------------------
// Name: PrintScores()
// Desc: Prints out all the scores in the array in ascending order.  The count
//		parameter determines the number of entries in the array.
//------------------------------------------------------------------------------
void PrintScores(float Scores[], int ScoreCount)
{
	int i;
	float SortedScores[MaxScoreCount];

	// sort the scores into ascending order and place in 'SortedScores' array
	SortAscending(Scores, SortedScores, ScoreCount);

	// output all scores from sorted array
	cout << "Printing Scores: " << endl;

	for (i = 0; i < ScoreCount; i++)
	{
		cout << SortedScores[i] << endl;
	}
}

//------------------------------------------------------------------------------
// Name: GetMean()
// Desc: The average (or mean) is a total of all values divided by the number
//		of values.  This function computes the average of all the scores in the
//		score array (Scores[]) using the amount of scores (ScoreCount) and
//		returns the result.
//------------------------------------------------------------------------------
float GetMean(float Scores[], int ScoreCount)
{
	int i;
	float Total = 0.0;

	// total all scores
	for (i = 0; i < ScoreCount; i++)
	{
		Total += Scores[i];
	}

	// divide by number of scores to get mean (average)
	return (Total / (float)ScoreCount);
}

//------------------------------------------------------------------------------
// Name: PrintAverage()
// Desc: This function prints the average score as gotten from 'GetMean()',
//		since the same averaging functionality must be utilized in the
//		calculation of the standard deviation.
//------------------------------------------------------------------------------
void PrintAverage(float Scores[], int ScoreCount)
{
	// display average
	cout << "Average score: " << GetMean(Scores, ScoreCount) << endl;
}

//------------------------------------------------------------------------------
// Name: PrintMedian()
// Desc: This calculates a SIMPLE median by picking out the middle value in the
//		list of scores, OR by averaging the two middle values in the case of an
//		even count.
//------------------------------------------------------------------------------
void PrintMedian(float Scores[], int ScoreCount)
{
	float SortedScores[MaxScoreCount];
	float Median;

	// sort the scores into ascending order and place in 'SortedScores' array
	SortAscending(Scores, SortedScores, ScoreCount);

	// check if count is odd by checking for a remainder in division by 2.  if
	// there IS a remainder then the count is ODD, otherwise it is even.
	if (ScoreCount % 2)
	{
		// Because of the way integer division works, the count divided by
		// two will be the whole parts without any fractional parts (the fract-
		// ional part is simply dropped, not rounded).  So, if the count is
		// odd, dividing it by two will yield the center index.

		// For example, if the count is 7, then 7 / 2 = 3:
		//
		// [0][1][2][3][4][5][6]
		//           ^
		Median = SortedScores[ScoreCount/2];
	}
	// otherwise if even, average the two middle values
	else
	{
		// The two middle values are the count divided by two and then that
		// index minus one (see above for integer division note).  The median
		// is the average of these two.
		Median = SortedScores[ScoreCount/2];
		Median += SortedScores[(ScoreCount/2) - 1];
		Median /= 2;
	}

	cout << "Median score: " << Median << endl;
}

//------------------------------------------------------------------------------
// Name: PrintStdDeviation()
// Desc: The standard deviation is the maximum amount from which the mean is
//		deviated from on average.  It is an "average of averages" of sorts.
//		This function gets the standard deviation of the array of scores and
//		prints the result.
//
//		The 'SumSquare' (sum of squares of the mean differences) is a double
//		because it can get fairly exponential in size.
//------------------------------------------------------------------------------
void PrintStdDeviation(float Scores[], int ScoreCount)
{
	int i;
	float Mean, MeanDiff, StdDeviation;
	double SumSquare = 0;

	// get the mean
	Mean = GetMean(Scores, ScoreCount);

	// loop through all scores
	for (i = 0; i < ScoreCount; i++)
	{
		// get the difference between the mean and the score subscript ... the
		// absolute value of this is not necessary since we multiply it by
		// itself below (which will always yield a positive).
		MeanDiff = Scores[i] - Mean;

		// add the difference squared (multiplied by itself) to our running sum
		SumSquare += (double)(MeanDiff * MeanDiff);
	}

	// divide the sum of squares by the number of scores minus one
	SumSquare /= (double)(ScoreCount - 1);

	// now get the square root of 'SumSquare' ... this is the standard deviation.
	// The 'sqrt()' function returns a double so it is casted explicitly to a
	// float.
	StdDeviation = (float) sqrt(SumSquare);

	// display standard deviation
	cout << "Standard Deviation: " << StdDeviation << endl;
}