//*[xreflist.h]*****************************************************************
//
// XrefList Class Definition
// By Neil C. Obremski
//
//******************************************************************************
#include "xreflist.h"		// XrefList Class Prototype
#include <ctype.h>			// tolower(), isalnum()
#include <string>			// strncpy(), strcmp()

using namespace std;

//_[XrefList::XrefList()]_______________________________________________________
//
// Default constructor.  Initializes all pointers to NULL.
//
XrefList::XrefList()
: m_pWords	(0)
, m_pSelWord(0)
, m_pSelLine(0)
{
}

//_[XrefList::~XrefList()]______________________________________________________
//
// Destructor.
//
XrefList::~XrefList()
{
	// Free all memory allocated by words.
	Clear();
}

//_[XrefList::Clear()]__________________________________________________________
//
// Clears the cross-referenced list of words, if any.
//
// Pre-Conditions: none.
// Post-Conditions: none.
//
void XrefList::Clear(void)
{
	WordNode	*pWord		= m_pWords;
	WordNode	*pNextWord	= 0;
	LineNode	*pLine		= 0;
	LineNode	*pNextLine	= 0;

	// Keep looping while words remain.
	while (pWord)
	{
		pNextWord = pWord->pNext;

		pLine = pWord->pLines;

		// Keep looping while lines remain.
		while (pLine)
		{
			pNextLine = pLine->pNext;

			delete pLine;

			pLine = pNextLine;

		} // while (pLine)

		delete pWord->sText;
		delete pWord;

		pWord = pNextWord;

	} // while (pWord)
}

//_[XrefList::ParseLine()]______________________________________________________
//
// Pre-conditions: none.
//
// Post-conditions: none.
//
bool XrefList::ParseLine(char *sLineText, int iLine)
{
	if (!sLineText)
		return false;

	if (iLine < 1)
		return false;

	// create buffer to store word (sWord) as well as a read pointer (rp) for
	// reading from 'sLineText' and a write pointer (wp) for writing to the
	// word buffer (sWord).
	char	*rp			= sLineText,
			*wp			= 0,
			*sWord		= new char [strlen(sLineText) + 1];

	if (!sWord)
		return false;

	// Return value for adding single word (AddWord()) and return value for
	// parsing entire line (fAccumReturn) which will be false if any call to
	// AddWord() fails, or true otherwise.
	bool	fReturn		= true,
			fAccumReturn= true;

	// loop until the NULL character is reached (i.e. zero)
	while (*rp)
	{
		// skip all non-alphanumeric characters
		if (!isalnum(*rp) && *rp)
			rp++;

		// copy all alphanumeric characters to 'sWord' buffer (using 'wp' pointer)
		wp = sWord;
		while (isalnum(*rp) && *rp)
		{
			*wp = *rp;
			rp++;
			wp++;
		}
		*wp = '\0';

		// Add the new-found word ... if the word already exists, the line will
		// simply be added rather than the whole word.
		fReturn = AddWord(sWord, iLine);
		if (false == fReturn)
			fAccumReturn = false;
	}

	return fAccumReturn;
}

//_[XrefList::AddWord()]________________________________________________________
//
// Pre-conditions: The word text (sText) must not be NULL or blank and the
// line number (iLine) must be at least one (1).  The word text must also not
// contain ANY non-alphanumeric characters.
//
// Post-conditions: none.
//
bool XrefList::AddWord(char *sText, int iLine)
{
	// Test for NULL or blank word text.
	if (!sText || !(*sText))
		return false;

	// Test for line less than one (1).
	if (iLine < 1)
		return false;

	// Verify that the word text is all alphanumeric characters.
	char *pt = sText;

	while (*pt)
	{
		// If a non-alphanumeric character is found, exit function with false!
		if (!isalnum(*pt))
			return false;
		pt++;
	}

	WordNode *pLastWord = 0;

	// Search the pre-existing linked list of words for the one specified.
	WordNode *pWord = m_pWords;

	while (pWord && strcmp(pWord->sText, sText))
	{
		pWord = pWord->pNext;
	}

	// If no pre-existing word by this text exists, create it
	if (!pWord)
	{
		pWord = new WordNode;

		// If allocation of word node fails, exit immediately!
		if (!pWord)
		{
			return false;
		}

		pWord->pLines = 0;
		pWord->sText = new char [strlen(sText) + 1];
		pWord->pNext = 0;

		// If allocation of word text fails, exit immediately!
		if (!pWord->sText)
		{
			delete pWord;
			return false;
		}

		// Copy passed word text into word node's text.
		strcpy(pWord->sText, sText);

		// Get pointer to first character in the string (for lower-casing).
		char *pt = pWord->sText;

		// Loop through all visible characters and make sure they are lower-
		// cased.
		while (*pt)
		{
			*pt = static_cast<char>(tolower(*pt));
			pt++;
		}

		// Add word to current linked list of words.
		if (m_pWords)
		{
			// Find the last word node.  This is apparent by the 'pNext' member
			// variable which will be NULL (0) for the last node (i.e. there IS
			// no next node).
			pLastWord = m_pWords;
			while (pLastWord->pNext)
				pLastWord = pLastWord->pNext;

			// Set this new word node as the last node.
			pLastWord->pNext = pWord;
		}
		// Or if this is the first word, set the linked list to it.
		else
		{
			m_pWords = pWord;
		}

	} // if (!pWord) ... created word

	// Add this new line number (or make sure it exists at least).
	return AddLineToWord_(pWord, iLine);
}

//_[XrefList::GetFirstWord()]___________________________________________________
//
// Pre-conditions: The text buffer (sBuf) must be non-NULL and the length of
// the buffer (iBufLen) must be at least two (2).
//
// Post-conditions: none.
//
bool XrefList::GetFirstWord(char *sBuf, int iBufLen)
{
	if (!sBuf)
		return false;

	if (iBufLen < 2)
		return false;

	// Select the first word.
	m_pSelWord = m_pWords;

	// Call 'GetWord()' to get the word selection into the buffer.  If there
	// is no selection this will simply return false.
	return GetWord(sBuf, iBufLen);
}

//_[XrefList::GetWord()]________________________________________________________
//
// Gets the currently selected word or returns false if there is no word
// selected.  To select the first word, use 'GetFirstWord()' and to move on
// through the list use 'ToNextWord()'.
//
// Pre-conditions: The text buffer (sBuf) must be non-NULL and the length of
// the buffer (iBufLen) must be at least two (2).
//
// Post-conditions: none.
//
bool XrefList::GetWord(char *sBuf, int iBufLen)
{
	// If a NULL or too-small buffer was passed in, return false
	if (!sBuf || iBufLen < 2)
		return false;

	// If a word is selected, copy its text into the specified buffer.
	if (m_pSelWord)
	{
		strncpy(sBuf, m_pSelWord->sText, iBufLen);
		return true;
	}

	// If there is no selection, "blank" the buffer and return false.
	*sBuf = '\0';
	return false;
}

//_[XrefList::ToNextWord()]_____________________________________________________
//
// Pre-conditions: A word node (m_pSelWord) must be selected.
//
// Post-conditions: A word node must have been moved to.
//
bool XrefList::ToNextWord(void)
{
	// If no word is selected, return false.
	if (!m_pSelWord)
		return false;

	// Otherwise move to the next word node.
	m_pSelWord = m_pSelWord->pNext;

	// If there is a new next node return true.
	if (m_pSelWord)
		return true;

	// Otherwise return false.
	return false;
}

//_[XrefList::GetLineCount()]___________________________________________________
//
// Returns the number of lines associated with the currently selected word.
//
// Pre-conditions: A word must be selected using 'GetFirstWord()' or 'GetWord()'.
//
// Post-conditions: none.
//
int XrefList::GetLineCount(void)
{
	if (!m_pSelWord)
		return -1;

	int			iCount	= 0;
	LineNode	*pLine	= m_pSelWord->pLines;

	// Loop through each line node in the linked list and increment 'iCount' for
	// each one found.
	while (pLine)
	{
		pLine = pLine->pNext;
		iCount++;
	}

	// Return 'iCount' which will be the number of line nodes associated with
	// the currently selected word (m_pSelWord).
	return iCount;
}

//_[XrefList::GetFirstLine()]___________________________________________________
//
// Pre-conditions: There must be word nodes available (if one isn't selected)
// and the selected word must have line nodes.
//
// Post-conditions: none.
//
int XrefList::GetFirstLine(void)
{
	// If a word is selected, return its first line value.
	if (m_pSelWord)
	{
		m_pSelLine = m_pSelWord->pLines;
		return GetLine();
	}

	// Otherwise try to select the first word and return its first line value
	m_pSelWord = m_pWords;

	if (m_pSelWord)
	{
		m_pSelLine = m_pSelWord->pLines;
		return GetLine();
	}

	// If there are no words then return '-1' for error!
	return -1;
}

//_[XrefList::GetLine()]________________________________________________________
//
// Pre-conditions: A line node must be selected (from 'GetFirstLine()' or
// 'ToNextLine()').
//
// Post-conditions: none.
//
int XrefList::GetLine(void)
{
	// If a line is selected, return its value.
	if (m_pSelLine)
		return m_pSelLine->iLine;

	// Otherwise return '-1' if no line is selected.
	return -1;
}

//_[XrefList::ToNextLine()]_____________________________________________________
//
// Pre-conditions: none.
//
// Post-conditions: none.
//
bool XrefList::ToNextLine(void)
{
	// If no line is selected return false.
	if (!m_pSelLine)
		return false;

	// Otherwise move to the next line node and return true.
	m_pSelLine = m_pSelLine->pNext;

	return true;
}

//_[XrefList::SortAscending()]__________________________________________________
//
// Slow bubble sort.  Sorts all words so they are in ascending order in the
// linked list.
//
// Pre-conditions: There must be at least two words in the list to sort.
// Post-conditions: none.
//
bool XrefList::SortAscending(void)
{
	WordNode	*pWord	= m_pWords,
				*pHold	= 0,
				*pPrev	= 0;
	int			i		= 0,
				iCount	= 0;

	// Check word list.
	if (!m_pWords)
		return false;

	// Check for first "pair" (must be at least two words in list).
	if (!pWord->pNext)
		return false;

	// Pass over word list a number of times to perform bubble sort.
	do
	{
		// Continue to loop as long as there is a PAIR ('pWord' and
		// 'pWord->pNext') of nodes to compare and possibly switch.
		while (pWord->pNext)
		{
			// If 'pWord' should go AFTER its next-door neighor, they are switched.
			// The 'pHold' variable is used to hold the original value of 'pWord'
			// while it is set to its neighbor.
			if (strcmp(pWord->sText, pWord->pNext->sText) > 0)
			{
				pHold = pWord;

				pWord = pWord->pNext;

				pHold->pNext = pWord->pNext;

				pWord->pNext = pHold;

				// If there was a previous word, set it's next to new first.
				if (pPrev)
					pPrev->pNext = pWord;
				// Otherwise keep head of list intact.
				else
					m_pWords = pWord;
			}

			// Set the current node as a previous one.
			pPrev = pWord;

			// Move to the next word node.
			pWord = pWord->pNext;

			// Use first iteration to get count of words so we can repeat this
			// until everything is properly sorted.
			if (0 == i)
			{
				iCount++;
			}

		} // while (pWord->pNext) ... single iteration through word list

		// reset iteration through word list
		pPrev = 0;
		pWord = m_pWords;

		// increment number of passes through list
		i++;

	} while (i < iCount); // pass through some word list iterations

	return true;
}

//_[XrefList::AddLineToWord_()]_________________________________________________
//
// Pre-conditions: none.
// Post-conditions: none.
//
bool XrefList::AddLineToWord_(WordNode *pWord, int iLine)
{
	// Word node check.
	if (!pWord)
		return false;

	// Line number check.
	if (iLine < 1)
		return false;

	// If there are already lines listed, add a new node or do nothing if a
	// node with the same line number is found.
	if (pWord->pLines)
	{
		LineNode *pLastLine = pWord->pLines;

		// Find the last line node in the linked list.  This would be the line
		// node whose 'pNext' member is NULL (i.e. there IS no next node).
		// Also exit the loop if a line node is found with the same line number,
		// because we don't allow duplicates.
		while (pLastLine->pNext && pLastLine->iLine != iLine)
			pLastLine = pLastLine->pNext;

		// Only add a new line node if it does not have a duplicate line number.
		if (pLastLine->iLine != iLine)
		{
			pLastLine->pNext = new LineNode;

			// If allocation of the line node fails, exit and return false!
			if (!pLastLine->pNext)
				return false;

			pLastLine->pNext->pNext = 0;
			pLastLine->pNext->iLine = iLine;
		}
	}
	// Otherwise if this is the first line node ...
	else
	{
		pWord->pLines = new LineNode;

		// If allocation of the line node fails, exit and return false!
		if (!pWord->pLines)
			return false;

		pWord->pLines->pNext = 0;
		pWord->pLines->iLine = iLine;
	}

	return true;
}